#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "bitstream.h"

#define MAGIC1 'B'
#define MAGIC2 'S'
#define MAGIC3 'D'
#define MAGIC4 'M'

#define CHECK_BSP(bsp)\
  assert(bsp != NULL);\
  assert(bsp->fp != NULL);\
  assert(bsp->phase >= 0 && bsp->phase < 8);

struct _bitstream
{
  FILE *fp;
  int phase;
  char buffer;
};

BITSTREAM *bitstream_wopen(FILE *fp)
{
  BITSTREAM *bsp;
  if (fp==NULL) return NULL;
  bsp=malloc(sizeof(BITSTREAM));
  bsp->fp = fp;
  bsp->phase = 0;
  bsp->buffer = 0;
  putc(MAGIC1, bsp->fp);
  putc(MAGIC2, bsp->fp);
  return bsp;
}

int bitstream_wclose(BITSTREAM *bsp)
{
  CHECK_BSP(bsp);
  if (bsp->phase > 0)
    putc(bsp->buffer << (8-bsp->phase), bsp->fp);
  putc(MAGIC3, bsp->fp);
  putc(MAGIC4, bsp->fp);
  return fclose(bsp->fp);
}

void bitstream_wbit(BITSTREAM *bsp, int bit)
{
  CHECK_BSP(bsp);
  bsp->buffer = (bsp->buffer << 1) | (bit ? 1 : 0);
  if (++bsp->phase == 8)
  {
    bsp->phase=0;
    putc(bsp->buffer, bsp->fp);
  }
}

BITSTREAM *bitstream_ropen(FILE *fp)
{
  BITSTREAM *bsp;
  if (fp==NULL) return NULL;
  if (getc(fp)!=MAGIC1 || getc(fp)!=MAGIC2)
    return NULL;
  bsp=malloc(sizeof(BITSTREAM));
  bsp->fp = fp;
  bsp->phase = 0;
  bsp->buffer = 0;
  return bsp;
}

int bitstream_rbit(BITSTREAM *bsp)
{
  int output;
  CHECK_BSP(bsp);
  if (bsp->phase == 0)
  {
    bsp->phase=8;
    bsp->buffer = getc(bsp->fp);
  }
  bsp->phase--;
  output = (bsp->buffer >> 7) & 1;
  bsp->buffer <<= 1;
  return output;
}

int bitstream_rclose(BITSTREAM *bsp)
{
  int error;
  CHECK_BSP(bsp);
  error = (getc(bsp->fp) != MAGIC3) || (getc(bsp->fp) != MAGIC4);
  return (fclose(bsp->fp) || error);
}


